@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -55,6 +55,7 @@
 #include "control.h"
 #include "sdpd.h"
 #include "glib-helper.h"
+#include "btio.h"
 #include "dbus-common.h"
 
 #define AVCTP_PSM 23
@@ -69,15 +70,11 @@
 #define AVCTP_PACKET_CONTINUE	2
 #define AVCTP_PACKET_END	3
 
-#define AVRCP_REJPACKET_LEN     0x01
-#define AVRCP_REJPACKET_CODE    0x00
-
 /* ctype entries */
 #define CTYPE_CONTROL		0x0
 #define CTYPE_STATUS		0x1
 #define CTYPE_ACCEPTED		0x9
 #define CTYPE_STABLE		0xC
-#define CTYPE_REJECT            0xA
 
 /* opcodes */
 #define OP_UNITINFO		0x30
@@ -168,9 +165,8 @@ struct avctp {
 
 	int uinput;
 
-	int sock;
-
-	guint io;
+	GIOChannel *io;
+	guint io_id;
 
 	uint16_t mtu;
 };
@@ -187,7 +183,7 @@ static sdp_record_t *avrcp_ct_record()
 	sdp_list_t *aproto, *proto[2];
 	sdp_record_t *record;
 	sdp_data_t *psm, *version, *features;
-	uint16_t lp = AVCTP_PSM, ver = 0x0103, feat = 0x000f;
+	uint16_t lp = AVCTP_PSM, ver = 0x0100, feat = 0x000f;
 
 	record = sdp_record_alloc();
 	if (!record)
@@ -250,7 +246,7 @@ static sdp_record_t *avrcp_tg_record()
 	sdp_list_t *aproto, *proto[2];
 	sdp_record_t *record;
 	sdp_data_t *psm, *version, *features;
-	uint16_t lp = AVCTP_PSM, ver = 0x0103, feat = 0x000f;
+	uint16_t lp = AVCTP_PSM, ver = 0x0100, feat = 0x000f;
 
 	record = sdp_record_alloc();
 	if (!record)
@@ -335,7 +331,6 @@ static struct avctp *avctp_get(const bdaddr_t *src, const bdaddr_t *dst)
 	session = g_new0(struct avctp, 1);
 
 	session->uinput = -1;
-	session->sock = -1;
 	bacpy(&session->src, src);
 	bacpy(&session->dst, dst);
 
@@ -432,10 +427,13 @@ static void avctp_unref(struct avctp *session)
 					DBUS_TYPE_BOOLEAN, &value);
 	}
 
-	if (session->sock >= 0)
-		close(session->sock);
-	if (session->io)
-		g_source_remove(session->io);
+	if (session->io) {
+		g_io_channel_shutdown(session->io, TRUE, NULL);
+		g_io_channel_unref(session->io);
+	}
+
+	if (session->io_id)
+		g_source_remove(session->io_id);
 
 	if (session->dev)
 		session->dev->control->session = NULL;
@@ -448,7 +446,6 @@ static void avctp_unref(struct avctp *session)
 	g_free(session);
 }
 
-
 static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 				gpointer data)
 {
@@ -456,18 +453,20 @@ static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 	unsigned char buf[1024], *operands;
 	struct avctp_header *avctp;
 	struct avrcp_header *avrcp;
-	int ret, packet_size, operand_count;
+	int ret, packet_size, operand_count, sock;
 
 	if (!(cond | G_IO_IN))
 		goto failed;
 
-	ret = read(session->sock, buf, sizeof(buf));
+	sock = g_io_channel_unix_get_fd(session->io);
+
+	ret = read(sock, buf, sizeof(buf));
 	if (ret <= 0)
 		goto failed;
 
 	debug("Got %d bytes of data for AVCTP session %p", ret, session);
 
-	if (ret < sizeof(struct avctp_header)) {
+	if ((unsigned int) ret < sizeof(struct avctp_header)) {
 		error("Too small AVCTP packet");
 		goto failed;
 	}
@@ -482,7 +481,7 @@ static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 			avctp->cr, avctp->ipid, ntohs(avctp->pid));
 
 	ret -= sizeof(struct avctp_header);
-	if (ret < sizeof(struct avrcp_header)) {
+	if ((unsigned int) ret < sizeof(struct avrcp_header)) {
 		error("Too small AVRCP packet");
 		goto failed;
 	}
@@ -509,10 +508,10 @@ static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 		handle_panel_passthrough(session, operands, operand_count);
 		avctp->cr = AVCTP_RESPONSE;
 		avrcp->code = CTYPE_ACCEPTED;
-		ret = write(session->sock, buf, packet_size);
+		ret = write(sock, buf, packet_size);
 	}
 
-	else if (avctp->packet_type == AVCTP_PACKET_SINGLE &&
+	if (avctp->packet_type == AVCTP_PACKET_SINGLE &&
 			avctp->cr == AVCTP_COMMAND &&
 			avctp->pid == htons(AV_REMOTE_SVCLASS_ID) &&
 			avrcp->code == CTYPE_STATUS &&
@@ -522,19 +521,7 @@ static gboolean session_cb(GIOChannel *chan, GIOCondition cond,
 		avrcp->code = CTYPE_STABLE;
 		debug("reply to %s", avrcp->opcode == OP_UNITINFO ?
 				"OP_UNITINFO" : "OP_SUBUNITINFO");
-		ret = write(session->sock, buf, packet_size);
-	}
-	else if(avctp->packet_type == AVCTP_PACKET_SINGLE)   
-	{
-		debug("AVRCP : INVALID PDU-ID received");
-		avctp->cr = AVCTP_RESPONSE;
-		avctp->ipid = 1;
-		avrcp->code = CTYPE_REJECT;
-		buf[11] = 0x00;  
-		buf[12] = AVRCP_REJPACKET_LEN;
-		buf[13] = AVRCP_REJPACKET_CODE;  
-		
-		ret = write(session->sock, buf, packet_size);
+		ret = write(sock, buf, packet_size);
 	}
 
 	return TRUE;
@@ -619,20 +606,45 @@ static void init_uinput(struct avctp *session)
 		debug("AVRCP: uinput initialized for %s", address);
 }
 
-static gboolean avctp_connect_session(struct avctp *session)
+static void avctp_connect_cb(GIOChannel *chan, GError *err, gpointer data)
 {
-	GIOChannel *io;
+	struct control *control = data;
+	struct avctp *session = control->session;
+	char address[18];
+	uint16_t imtu;
+	GError *gerr = NULL;
 	gboolean value;
 
-	session->dev = manager_find_device(&session->dst, NULL, FALSE);
-	if (!session->dev) {
-		error("Connecting audio device not known (SDP not completed)");
-		return FALSE;
+	if (!session) {
+		debug("avctp_connect_cb: session removed while connecting");
+		if (!err)
+			g_io_channel_shutdown(chan, TRUE, NULL);
+		g_io_channel_unref(chan);
+		return;
 	}
 
-	session->state = AVCTP_STATE_CONNECTED;
+	if (err) {
+		avctp_unref(session);
+		error("%s", err->message);
+		return;
+	}
 
-	session->dev->control->session = session;
+	bt_io_get(chan, BT_IO_L2CAP, &gerr,
+			BT_IO_OPT_DEST, &address,
+			BT_IO_OPT_IMTU, &imtu,
+			BT_IO_OPT_INVALID);
+	if (gerr) {
+		avctp_unref(session);
+		error("%s", gerr->message);
+		g_error_free(gerr);
+		g_io_channel_shutdown(chan, TRUE, NULL);
+		return;
+	}
+
+	debug("AVCTP: connected to %s", address);
+
+	if (!session->io)
+		session->io = g_io_channel_ref(chan);
 
 	init_uinput(session);
 
@@ -644,60 +656,66 @@ static gboolean avctp_connect_session(struct avctp *session)
 				AUDIO_CONTROL_INTERFACE, "Connected",
 				DBUS_TYPE_BOOLEAN, &value);
 
-	if (session->io)
-		g_source_remove(session->io);
-
-	io = g_io_channel_unix_new(session->sock);
-	session->io = g_io_add_watch(io,
-			G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-			(GIOFunc) session_cb, session);
-	g_io_channel_unref(io);
-
-	return TRUE;
+	session->state = AVCTP_STATE_CONNECTED;
+	session->mtu = imtu;
+	session->io_id = g_io_add_watch(chan,
+				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
+				(GIOFunc) session_cb, session);
 }
 
 static void auth_cb(DBusError *derr, void *user_data)
 {
-	struct avctp *session = user_data;
+	struct control *control = user_data;
+	struct avctp *session = control->session;
+	GError *err = NULL;
 
 	if (derr && dbus_error_is_set(derr)) {
 		error("Access denied: %s", derr->message);
-
 		avctp_unref(session);
 		return;
 	}
 
-	avctp_connect_session(session);
+	if (!bt_io_accept(session->io, avctp_connect_cb, control,
+								NULL, &err)) {
+		error("bt_io_accept: %s", err->message);
+		g_error_free(err);
+		avctp_unref(session);
+	}
 }
 
-static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-				const bdaddr_t *dst, gpointer data)
+static void avctp_confirm_cb(GIOChannel *chan, gpointer data)
 {
-	socklen_t size;
-	struct l2cap_options l2o;
 	struct avctp *session;
-	GIOCondition flags = G_IO_ERR | G_IO_HUP | G_IO_NVAL;
 	struct audio_device *dev;
 	char address[18];
+	bdaddr_t src, dst;
+	GError *err = NULL;
 
-	if (err < 0) {
-		error("AVCTP server socket: %s (%d)", strerror(-err), -err);
+	bt_io_get(chan, BT_IO_L2CAP, &err,
+			BT_IO_OPT_SOURCE_BDADDR, &src,
+			BT_IO_OPT_DEST_BDADDR, &dst,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		error("%s", err->message);
+		g_error_free(err);
+		g_io_channel_shutdown(chan, TRUE, NULL);
 		return;
 	}
 
-	session = avctp_get(src, dst);
+	session = avctp_get(&src, &dst);
 
 	if (!session) {
 		error("Unable to create new AVCTP session");
 		goto drop;
 	}
 
-	if (session->sock >= 0) {
+	if (session->io) {
 		error("Refusing unexpected connect from %s", address);
 		goto drop;
 	}
 
-	dev = manager_get_device(src, dst);
+	dev = manager_get_device(&src, &dst);
 	if (!dev) {
 		error("Unable to get audio device object for %s", address);
 		goto drop;
@@ -706,125 +724,59 @@ static void avctp_server_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	if (!dev->control)
 		dev->control = control_init(dev);
 
+	if (!dev->control->session)
+		dev->control->session = session;
+
+	if (!session->dev)
+		session->dev = dev;
+
+	device_remove_control_timer(dev);
+
 	session->state = AVCTP_STATE_CONNECTING;
-	session->sock = g_io_channel_unix_get_fd(chan);
+	session->io = g_io_channel_ref(chan);
 
-	memset(&l2o, 0, sizeof(l2o));
-	size = sizeof(l2o);
-	if (getsockopt(session->sock, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &size) < 0) {
-		err = errno;
-		error("getsockopt(L2CAP_OPTIONS): %s (%d)", strerror(err),
-				err);
-		goto drop;
-	}
-
-	session->mtu = l2o.imtu;
-
-	session->io = g_io_add_watch(chan, flags, (GIOFunc) session_cb,
-				session);
-	g_io_channel_unref(chan);
-
-	if (avdtp_is_connected(src, dst))
-		goto proceed;
-
-	err = btd_request_authorization(src, dst, AVRCP_TARGET_UUID,
-				auth_cb, session);
-	if (err < 0)
-		goto drop;
-
-	return;
-
-proceed:
-	if (!avctp_connect_session(session))
+	if (avdtp_is_connected(&src, &dst)) {
+		if (!bt_io_accept(chan, avctp_connect_cb, dev->control,
+								NULL, NULL))
+			goto drop;
+	} else if (btd_request_authorization(&src, &dst,
+				AVRCP_TARGET_UUID, auth_cb, dev->control) < 0)
 		goto drop;
 
 	return;
 
 drop:
-	g_io_channel_close(chan);
+	if (!session->io)
+		g_io_channel_shutdown(chan, TRUE, NULL);
 	avctp_unref(session);
 }
 
 static GIOChannel *avctp_server_socket(const bdaddr_t *src, gboolean master)
 {
+	GError *err = NULL;
 	GIOChannel *io;
-	int lm = L2CAP_LM_AUTH | L2CAP_LM_ENCRYPT;
 
-	if (master)
-		lm |= L2CAP_LM_MASTER;
-
-	io = bt_l2cap_listen(src, AVCTP_PSM, 0, lm, avctp_server_cb,
-				NULL);
+	io = bt_io_listen(BT_IO_L2CAP, NULL, avctp_confirm_cb, NULL,
+				NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, src,
+				BT_IO_OPT_PSM, AVCTP_PSM,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_MEDIUM,
+				BT_IO_OPT_MASTER, master,
+				BT_IO_OPT_INVALID);
 	if (!io) {
-		error("Unable to allocate new io channel");
-		return NULL;
+		error("%s", err->message);
+		g_error_free(err);
 	}
 
 	return io;
 }
 
-static void avctp_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-			const bdaddr_t *dst, gpointer data)
-{
-	struct control *control = data;
-	struct avctp *session = control->session;
-	struct l2cap_options l2o;
-	socklen_t len;
-	int sk;
-	char address[18];
-
-	if (!session) {
-		debug("avctp_connect_cb: session removed while connecting");
-		g_io_channel_close(chan);
-		g_io_channel_unref(chan);
-		return;
-	}
-
-	ba2str(&session->dst, address);
-
-	if (err < 0) {
-		avctp_unref(session);
-		error("AVCTP connect(%s): %s (%d)", address, strerror(-err),
-				-err);
-		return;
-	}
-
-	debug("AVCTP: connected to %s", address);
-
-	g_io_channel_set_close_on_unref(chan, FALSE);
-	sk = g_io_channel_unix_get_fd(chan);
-	session->sock = sk;
-
-	memset(&l2o, 0, sizeof(l2o));
-	len = sizeof(l2o);
-	if (getsockopt(sk, SOL_L2CAP, L2CAP_OPTIONS, &l2o, &len) < 0) {
-		err = errno;
-		g_io_channel_unref(chan);
-		avctp_unref(session);
-		error("getsockopt(L2CAP_OPTIONS): %s (%d)", strerror(err),
-				err);
-		return;
-	}
-
-	init_uinput(session);
-
-	g_dbus_emit_signal(session->dev->conn, session->dev->path,
-				AUDIO_CONTROL_INTERFACE, "Connected",
-				DBUS_TYPE_INVALID);
-
-	session->state = AVCTP_STATE_CONNECTED;
-	session->mtu = l2o.imtu;
-	session->io = g_io_add_watch(chan,
-				G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
-				(GIOFunc) session_cb, session);
-	g_io_channel_unref(chan);
-}
-
 gboolean avrcp_connect(struct audio_device *dev)
 {
 	struct control *control = dev->control;
 	struct avctp *session;
-	int err;
+	GError *err = NULL;
+	GIOChannel *io;
 
 	if (control->session)
 		return TRUE;
@@ -835,17 +787,25 @@ gboolean avrcp_connect(struct audio_device *dev)
 		return FALSE;
 	}
 
+	device_remove_control_timer(dev);
+
 	session->dev = dev;
 	session->state = AVCTP_STATE_CONNECTING;
 
-	err = bt_l2cap_connect(&dev->src, &dev->dst, AVCTP_PSM, 0,
-				avctp_connect_cb, control);
-	if (err < 0) {
+	io = bt_io_connect(BT_IO_L2CAP, avctp_connect_cb, control, NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+				BT_IO_OPT_DEST_BDADDR, &dev->dst,
+				BT_IO_OPT_PSM, AVCTP_PSM,
+				BT_IO_OPT_INVALID);
+	if (err) {
 		avctp_unref(session);
-		error("Connect failed. %s(%d)", strerror(-err), -err);
+		error("%s", err->message);
+		g_error_free(err);
 		return FALSE;
 	}
 
+	g_io_channel_unref(io);
+
 	control->session = session;
 
 	return TRUE;
@@ -955,7 +915,7 @@ void avrcp_unregister(const bdaddr_t *src)
 	remove_record_from_server(server->ct_record_id);
 	remove_record_from_server(server->tg_record_id);
 
-	g_io_channel_close(server->io);
+	g_io_channel_shutdown(server->io, TRUE, NULL);
 	g_io_channel_unref(server->io);
 	g_free(server);
 
