@@ -2,7 +2,7 @@
  *
  *  BlueZ - Bluetooth protocol stack for Linux
  *
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This library is free software; you can redistribute it and/or
@@ -53,16 +53,16 @@ GST_DEBUG_CATEGORY_STATIC(avdtp_sink_debug);
 #define DEFAULT_AUTOCONNECT TRUE
 
 #define GST_AVDTP_SINK_MUTEX_LOCK(s) G_STMT_START {	\
-		g_mutex_lock (s->sink_lock);		\
+		g_mutex_lock(s->sink_lock);		\
 	} G_STMT_END
 
 #define GST_AVDTP_SINK_MUTEX_UNLOCK(s) G_STMT_START {	\
-		g_mutex_unlock (s->sink_lock);		\
+		g_mutex_unlock(s->sink_lock);		\
 	} G_STMT_END
 
 
 struct bluetooth_data {
-	struct bt_getcapabilities_rsp caps; /* Bluetooth device caps */
+	struct bt_get_capabilities_rsp *caps; /* Bluetooth device caps */
 	guint link_mtu;
 
 	gchar buffer[BUFFER_SIZE];	/* Codec transfer buffer */
@@ -113,7 +113,7 @@ static GIOError gst_avdtp_sink_audioservice_send(GstAvdtpSink *self,
 static GIOError gst_avdtp_sink_audioservice_expect(
 				GstAvdtpSink *self,
 				bt_audio_msg_header_t *outmsg,
-				int expected_type);
+				guint8 expected_name);
 
 
 static void gst_avdtp_sink_base_init(gpointer g_class)
@@ -144,8 +144,7 @@ static gboolean gst_avdtp_sink_stop(GstBaseSink *basesink)
 	}
 
 	if (self->stream) {
-		g_io_channel_flush(self->stream, NULL);
-		g_io_channel_close(self->stream);
+		g_io_channel_shutdown(self->stream, TRUE, NULL);
 		g_io_channel_unref(self->stream);
 		self->stream = NULL;
 	}
@@ -243,16 +242,38 @@ static gint gst_avdtp_sink_bluetooth_recvmsg_fd(GstAvdtpSink *sink)
 	return 0;
 }
 
+static codec_capabilities_t *gst_avdtp_find_caps(GstAvdtpSink *sink,
+						uint8_t codec_type)
+{
+	struct bt_get_capabilities_rsp *rsp = sink->data->caps;
+	codec_capabilities_t *codec = (void *) rsp->data;
+	int bytes_left = rsp->h.length - sizeof(*rsp);
+
+	while (bytes_left > 0) {
+		if (codec->type == codec_type)
+			break;
+
+		bytes_left -= codec->length;
+		codec = (void *) codec + codec->length;
+	}
+
+	if (bytes_left <= 0)
+		return NULL;
+
+	return codec;
+}
+
 static gboolean gst_avdtp_sink_init_sbc_pkt_conf(GstAvdtpSink *sink,
 					GstCaps *caps,
 					sbc_capabilities_t *pkt)
 {
-	sbc_capabilities_t *cfg = &sink->data->caps.sbc_capabilities;
+	sbc_capabilities_t *cfg;
 	const GValue *value = NULL;
 	const char *pref, *name;
 	gint rate, subbands, blocks;
 	GstStructure *structure = gst_caps_get_structure(caps, 0);
 
+	cfg = (void *) gst_avdtp_find_caps(sink, BT_A2DP_CODEC_SBC);
 	name = gst_structure_get_name(structure);
 
 	if (!(IS_SBC(name))) {
@@ -595,6 +616,9 @@ static GstStructure *gst_avdtp_sink_parse_mpeg_caps(
 	gboolean valid_layer = FALSE;
 	gboolean mono, stereo;
 
+	if (!mpeg)
+		return NULL;
+
 	GST_LOG_OBJECT(self, "parsing mpeg caps");
 
 	structure = gst_structure_empty_new("audio/mpeg");
@@ -709,14 +733,17 @@ static GstStructure *gst_avdtp_sink_parse_mpeg_caps(
 
 static gboolean gst_avdtp_sink_update_caps(GstAvdtpSink *self)
 {
-	sbc_capabilities_t *sbc = &self->data->caps.sbc_capabilities;
-	mpeg_capabilities_t *mpeg = &self->data->caps.mpeg_capabilities;
+	sbc_capabilities_t *sbc;
+	mpeg_capabilities_t *mpeg;
 	GstStructure *sbc_structure;
 	GstStructure *mpeg_structure;
 	gchar *tmp;
 
 	GST_LOG_OBJECT(self, "updating device caps");
 
+	sbc = (void *) gst_avdtp_find_caps(self, BT_A2DP_CODEC_SBC);
+	mpeg = (void *) gst_avdtp_find_caps(self, BT_A2DP_CODEC_MPEG12);
+
 	sbc_structure = gst_avdtp_sink_parse_sbc_caps(self, sbc);
 	mpeg_structure = gst_avdtp_sink_parse_mpeg_caps(self, mpeg);
 
@@ -735,14 +762,17 @@ static gboolean gst_avdtp_sink_update_caps(GstAvdtpSink *self)
 
 static gboolean gst_avdtp_sink_get_capabilities(GstAvdtpSink *self)
 {
-	gchar *buf[BT_AUDIO_IPC_PACKET_SIZE];
-	struct bt_getcapabilities_req *req = (void *) buf;
-	struct bt_getcapabilities_rsp *rsp = (void *) buf;
+	gchar *buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_get_capabilities_req *req = (void *) buf;
+	struct bt_get_capabilities_rsp *rsp = (void *) buf;
 	GIOError io_error;
 
-	memset(req, 0, BT_AUDIO_IPC_PACKET_SIZE);
+	memset(req, 0, BT_SUGGESTED_BUFFER_SIZE);
+
+	req->h.type = BT_REQUEST;
+	req->h.name = BT_GET_CAPABILITIES;
+	req->h.length = sizeof(*req);
 
-	req->h.msg_type = BT_GETCAPABILITIES_REQ;
 	if (self->device == NULL)
 		return FALSE;
 	strncpy(req->device, self->device, 18);
@@ -755,21 +785,16 @@ static gboolean gst_avdtp_sink_get_capabilities(GstAvdtpSink *self)
 		return FALSE;
 	}
 
+	rsp->h.length = 0;
 	io_error = gst_avdtp_sink_audioservice_expect(self,
-			&rsp->rsp_h.msg_h, BT_GETCAPABILITIES_RSP);
+			&rsp->h, BT_GET_CAPABILITIES);
 	if (io_error != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error while getting device caps");
 		return FALSE;
 	}
 
-	if (rsp->rsp_h.posix_errno != 0) {
-		GST_ERROR_OBJECT(self, "BT_GETCAPABILITIES failed : %s(%d)",
-					strerror(rsp->rsp_h.posix_errno),
-					rsp->rsp_h.posix_errno);
-		return FALSE;
-	}
-
-	memcpy(&self->data->caps, rsp, sizeof(*rsp));
+	self->data->caps = g_malloc0(rsp->h.length);
+	memcpy(self->data->caps, rsp, rsp->h.length);
 	if (!gst_avdtp_sink_update_caps(self)) {
 		GST_WARNING_OBJECT(self, "failed to update capabilities");
 		return FALSE;
@@ -782,9 +807,9 @@ static gint gst_avdtp_sink_get_channel_mode(const gchar *mode)
 {
 	if (strcmp(mode, "stereo") == 0)
 		return BT_A2DP_CHANNEL_MODE_STEREO;
-	else if (strcmp(mode, "joint") == 0)
+	else if (strcmp(mode, "joint-stereo") == 0)
 		return BT_A2DP_CHANNEL_MODE_JOINT_STEREO;
-	else if (strcmp(mode, "dual") == 0)
+	else if (strcmp(mode, "dual-channel") == 0)
 		return BT_A2DP_CHANNEL_MODE_DUAL_CHANNEL;
 	else if (strcmp(mode, "mono") == 0)
 		return BT_A2DP_CHANNEL_MODE_MONO;
@@ -793,7 +818,7 @@ static gint gst_avdtp_sink_get_channel_mode(const gchar *mode)
 }
 
 static void gst_avdtp_sink_tag(const GstTagList *taglist,
-			const gchar* tag, gpointer user_data)
+			const gchar *tag, gpointer user_data)
 {
 	gboolean crc;
 	gchar *channel_mode = NULL;
@@ -888,16 +913,16 @@ failed:
 
 static gboolean gst_avdtp_sink_stream_start(GstAvdtpSink *self)
 {
-	gchar buf[BT_AUDIO_IPC_PACKET_SIZE];
-	struct bt_streamstart_req *req = (void *) buf;
-	struct bt_streamstart_rsp *rsp = (void *) buf;
-	struct bt_streamfd_ind *ind = (void*) buf;
+	gchar buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_start_stream_req *req = (void *) buf;
+	struct bt_start_stream_rsp *rsp = (void *) buf;
+	struct bt_new_stream_ind *ind = (void *) buf;
 	GIOError io_error;
 
-	GST_DEBUG_OBJECT(self, "stream start");
-
-	memset (req, 0, sizeof(buf));
-	req->h.msg_type = BT_STREAMSTART_REQ;
+	memset(req, 0, sizeof(buf));
+	req->h.type = BT_REQUEST;
+	req->h.name = BT_START_STREAM;
+	req->h.length = sizeof(*req);
 
 	io_error = gst_avdtp_sink_audioservice_send(self, &req->h);
 	if (io_error != G_IO_ERROR_NONE) {
@@ -906,27 +931,18 @@ static gboolean gst_avdtp_sink_stream_start(GstAvdtpSink *self)
 		return FALSE;
 	}
 
-	GST_DEBUG_OBJECT(self, "stream start packet sent");
-
+	rsp->h.length = sizeof(*rsp);
 	io_error = gst_avdtp_sink_audioservice_expect(self,
-			&rsp->rsp_h.msg_h, BT_STREAMSTART_RSP);
+			&rsp->h, BT_START_STREAM);
 	if (io_error != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error while stream "
 			"start confirmation");
 		return FALSE;
 	}
 
-	if (rsp->rsp_h.posix_errno != 0) {
-		GST_ERROR_OBJECT(self, "BT_STREAMSTART_RSP failed : %s(%d)",
-					strerror(rsp->rsp_h.posix_errno),
-					rsp->rsp_h.posix_errno);
-		return FALSE;
-	}
-
-	GST_DEBUG_OBJECT(self, "stream started");
-
+	ind->h.length = sizeof(*ind);
 	io_error = gst_avdtp_sink_audioservice_expect(self, &ind->h,
-			BT_STREAMFD_IND);
+			BT_NEW_STREAM);
 	if (io_error != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error while receiving "
 			"stream filedescriptor");
@@ -945,7 +961,7 @@ static gboolean gst_avdtp_sink_init_mp3_pkt_conf(
 {
 	const GValue *value = NULL;
 	gint rate, layer;
-	const gchar* name;
+	const gchar *name;
 	GstStructure *structure = gst_caps_get_structure(caps, 0);
 
 	name = gst_structure_get_name(structure);
@@ -1019,9 +1035,9 @@ static gboolean gst_avdtp_sink_init_mp3_pkt_conf(
 static gboolean gst_avdtp_sink_configure(GstAvdtpSink *self,
 			GstCaps *caps)
 {
-	gchar buf[BT_AUDIO_IPC_PACKET_SIZE];
-	struct bt_setconfiguration_req *req = (void *) buf;
-	struct bt_setconfiguration_rsp *rsp = (void *) buf;
+	gchar buf[BT_SUGGESTED_BUFFER_SIZE];
+	struct bt_set_configuration_req *req = (void *) buf;
+	struct bt_set_configuration_rsp *rsp = (void *) buf;
 	gboolean ret;
 	GIOError io_error;
 	gchar *temp;
@@ -1031,18 +1047,21 @@ static gboolean gst_avdtp_sink_configure(GstAvdtpSink *self,
 	GST_DEBUG_OBJECT(self, "configuring device with caps: %s", temp);
 	g_free(temp);
 
-	memset (req, 0, sizeof(buf));
-	req->h.msg_type = BT_SETCONFIGURATION_REQ;
+	memset(req, 0, sizeof(buf));
+	req->h.type = BT_REQUEST;
+	req->h.name = BT_SET_CONFIGURATION;
+	req->h.length = sizeof(*req);
+
 	req->access_mode = BT_CAPABILITIES_ACCESS_MODE_WRITE;
 	strncpy(req->device, self->device, 18);
 	structure = gst_caps_get_structure(caps, 0);
 
 	if (gst_structure_has_name(structure, "audio/x-sbc"))
 		ret = gst_avdtp_sink_init_sbc_pkt_conf(self, caps,
-				&req->sbc_capabilities);
+				(void *) &req->codec);
 	else if (gst_structure_has_name(structure, "audio/mpeg"))
 		ret = gst_avdtp_sink_init_mp3_pkt_conf(self, caps,
-				&req->mpeg_capabilities);
+				(void *) &req->codec);
 	else
 		ret = FALSE;
 
@@ -1052,6 +1071,7 @@ static gboolean gst_avdtp_sink_configure(GstAvdtpSink *self,
 		return FALSE;
 	}
 
+	req->h.length += req->codec.length - sizeof(req->codec);
 	io_error = gst_avdtp_sink_audioservice_send(self, &req->h);
 	if (io_error != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error ocurred while sending "
@@ -1059,26 +1079,16 @@ static gboolean gst_avdtp_sink_configure(GstAvdtpSink *self,
 		return FALSE;
 	}
 
-	GST_DEBUG_OBJECT(self, "configuration packet sent");
-
+	rsp->h.length = sizeof(*rsp);
 	io_error = gst_avdtp_sink_audioservice_expect(self,
-			&rsp->rsp_h.msg_h, BT_SETCONFIGURATION_RSP);
+			&rsp->h, BT_SET_CONFIGURATION);
 	if (io_error != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error while receiving device "
 					"confirmation");
 		return FALSE;
 	}
 
-	if (rsp->rsp_h.posix_errno != 0) {
-		GST_ERROR_OBJECT(self, "BT_SETCONFIGURATION_RSP failed : "
-					"%s(%d)",
-					strerror(rsp->rsp_h.posix_errno),
-					rsp->rsp_h.posix_errno);
-		return FALSE;
-	}
-
 	self->data->link_mtu = rsp->link_mtu;
-	GST_DEBUG_OBJECT(self, "configuration set");
 
 	return TRUE;
 }
@@ -1108,8 +1118,9 @@ static GstFlowReturn gst_avdtp_sink_render(GstBaseSink *basesink,
 	gsize ret;
 	GIOError err;
 
-	err = g_io_channel_write(self->stream, (gchar*)GST_BUFFER_DATA(buffer),
-			 (gsize)(GST_BUFFER_SIZE(buffer)), &ret);
+	err = g_io_channel_write(self->stream,
+				(gchar *) GST_BUFFER_DATA(buffer),
+				(gsize) (GST_BUFFER_SIZE(buffer)), &ret);
 
 	if (err != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error while writting to socket: %d %s",
@@ -1125,13 +1136,13 @@ static gboolean gst_avdtp_sink_unlock(GstBaseSink *basesink)
 	GstAvdtpSink *self = GST_AVDTP_SINK(basesink);
 
 	if (self->stream != NULL)
-		g_io_channel_flush (self->stream, NULL);
+		g_io_channel_flush(self->stream, NULL);
 
 	return TRUE;
 }
 
 static GstFlowReturn gst_avdtp_sink_buffer_alloc(GstBaseSink *basesink,
-				guint64 offset, guint size, GstCaps* caps,
+				guint64 offset, guint size, GstCaps *caps,
 				GstBuffer **buf)
 {
 	GstAvdtpSink *self = GST_AVDTP_SINK(basesink);
@@ -1219,13 +1230,22 @@ static GIOError gst_avdtp_sink_audioservice_send(
 {
 	GIOError error;
 	gsize written;
+	const char *type, *name;
+	uint16_t length;
 
-	error = g_io_channel_write(self->server, (const gchar*) msg,
-			BT_AUDIO_IPC_PACKET_SIZE, &written);
+	length = msg->length ? msg->length : BT_SUGGESTED_BUFFER_SIZE;
+
+	error = g_io_channel_write(self->server, (const gchar *) msg, length,
+								&written);
 	if (error != G_IO_ERROR_NONE)
 		GST_ERROR_OBJECT(self, "Error sending data to audio service:"
 			" %s(%d)", strerror(errno), errno);
 
+	type = bt_audio_strtype(msg->type);
+	name = bt_audio_strname(msg->name);
+
+	GST_DEBUG_OBJECT(self, "sent: %s -> %s", type, name);
+
 	return error;
 }
 
@@ -1235,30 +1255,53 @@ static GIOError gst_avdtp_sink_audioservice_recv(
 {
 	GIOError status;
 	gsize bytes_read;
-	const char *type;
+	const char *type, *name;
+	uint16_t length;
 
-	status = g_io_channel_read(self->server, (gchar*) inmsg,
-			BT_AUDIO_IPC_PACKET_SIZE, &bytes_read);
+	length = inmsg->length ? inmsg->length : BT_SUGGESTED_BUFFER_SIZE;
+
+	status = g_io_channel_read(self->server, (gchar *) inmsg, length,
+								&bytes_read);
 	if (status != G_IO_ERROR_NONE) {
 		GST_ERROR_OBJECT(self, "Error receiving data from "
 				"audio service");
 		return status;
 	}
 
-	type = bt_audio_strmsg(inmsg->msg_type);
+	type = bt_audio_strtype(inmsg->type);
 	if (!type) {
 		status = G_IO_ERROR_INVAL;
 		GST_ERROR_OBJECT(self, "Bogus message type %d "
 				"received from audio service",
-				inmsg->msg_type);
+				inmsg->type);
 	}
 
+	name = bt_audio_strname(inmsg->name);
+	if (!name) {
+		status = G_IO_ERROR_INVAL;
+		GST_ERROR_OBJECT(self, "Bogus message name %d "
+				"received from audio service",
+				inmsg->name);
+	}
+
+	if (inmsg->type == BT_ERROR) {
+		bt_audio_error_t *err = (void *) inmsg;
+		status = G_IO_ERROR_INVAL;
+		GST_ERROR_OBJECT(self, "%s failed : "
+					"%s(%d)",
+					name,
+					strerror(err->posix_errno),
+					err->posix_errno);
+	}
+
+	GST_DEBUG_OBJECT(self, "received: %s <- %s", type, name);
+
 	return status;
 }
 
 static GIOError gst_avdtp_sink_audioservice_expect(
 			GstAvdtpSink *self, bt_audio_msg_header_t *outmsg,
-			int expected_type)
+			guint8 expected_name)
 {
 	GIOError status;
 
@@ -1266,16 +1309,16 @@ static GIOError gst_avdtp_sink_audioservice_expect(
 	if (status != G_IO_ERROR_NONE)
 		return status;
 
-	if (outmsg->msg_type != expected_type)
+	if (outmsg->name != expected_name)
 		status = G_IO_ERROR_INVAL;
 
 	return status;
 }
 
-gboolean gst_avdtp_sink_plugin_init (GstPlugin * plugin)
+gboolean gst_avdtp_sink_plugin_init(GstPlugin *plugin)
 {
-	return gst_element_register (plugin, "avdtpsink",
-			GST_RANK_NONE, GST_TYPE_AVDTP_SINK);
+	return gst_element_register(plugin, "avdtpsink", GST_RANK_NONE,
+							GST_TYPE_AVDTP_SINK);
 }
 
 
@@ -1311,7 +1354,7 @@ guint gst_avdtp_sink_get_link_mtu(GstAvdtpSink *sink)
 	return sink->data->link_mtu;
 }
 
-void gst_avdtp_sink_set_device(GstAvdtpSink *self, const gchar* dev)
+void gst_avdtp_sink_set_device(GstAvdtpSink *self, const gchar *dev)
 {
 	if (self->device != NULL)
 		g_free(self->device);
@@ -1357,5 +1400,3 @@ void gst_avdtp_sink_set_channel_mode(GstAvdtpSink *self,
 		GST_WARNING_OBJECT(self, "Received invalid channel "
 				"mode: %s", mode);
 }
-
-
