@@ -58,6 +58,7 @@
 #include "telephony.h"
 #include "headset.h"
 #include "glib-helper.h"
+#include "btio.h"
 #include "dbus-common.h"
 #include "../src/adapter.h"
 #include "../src/device.h"
@@ -403,13 +404,16 @@ static void pending_connect_finalize(struct audio_device *dev)
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 
+	if (p == NULL)
+		return;
+
 	g_slist_foreach(p->callbacks, (GFunc) pending_connect_complete, dev);
 
 	g_slist_foreach(p->callbacks, (GFunc) g_free, NULL);
 	g_slist_free(p->callbacks);
 
 	if (p->io) {
-		g_io_channel_close(p->io);
+		g_io_channel_shutdown(p->io, TRUE, NULL);
 		g_io_channel_unref(p->io);
 	}
 
@@ -509,21 +513,21 @@ static gboolean ring_timer_cb(gpointer data)
 	return TRUE;
 }
 
-static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-			const bdaddr_t *dst, gpointer user_data)
+static void sco_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 {
 	int sk;
 	struct audio_device *dev = user_data;
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 
-	if (err < 0) {
-		error("connect(): %s (%d)", strerror(-err), -err);
+	if (err) {
+		error("%s", err->message);
 
-		if (p->msg)
+		if (p && p->msg)
 			error_connection_attempt_failed(dev->conn, p->msg, p->err);
 
 		pending_connect_finalize(dev);
+
 		if (hs->rfcomm)
 			headset_set_state(dev, HEADSET_STATE_CONNECTED);
 		else
@@ -537,16 +541,19 @@ static void sco_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	sk = g_io_channel_unix_get_fd(chan);
 
 	debug("SCO fd=%d", sk);
-	hs->sco = chan;
-	p->io = NULL;
+	hs->sco = g_io_channel_ref(chan);
 
-	if (p->msg) {
-		DBusMessage *reply = dbus_message_new_method_return(p->msg);
-		g_dbus_send_message(dev->conn, reply);
+	if (p) {
+		p->io = NULL;
+		if (p->msg) {
+			DBusMessage *reply;
+			reply = dbus_message_new_method_return(p->msg);
+			g_dbus_send_message(dev->conn, reply);
+		}
+
+		pending_connect_finalize(dev);
 	}
 
-	pending_connect_finalize(dev);
-
 	fcntl(sk, F_SETFL, 0);
 
 	headset_set_state(dev, HEADSET_STATE_PLAYING);
@@ -564,17 +571,24 @@ static int sco_connect(struct audio_device *dev, headset_stream_cb_t cb,
 			void *user_data, unsigned int *cb_id)
 {
 	struct headset *hs = dev->headset;
-	int err;
+	GError *err = NULL;
+	GIOChannel *io;
 
 	if (hs->state != HEADSET_STATE_CONNECTED)
 		return -EINVAL;
 
-	err = bt_sco_connect(&dev->src, &dev->dst, sco_connect_cb, dev);
-	if (err < 0) {
-		error("connect: %s (%d)", strerror(-err), -err);
-		return err;
+	io = bt_io_connect(BT_IO_SCO, sco_connect_cb, dev, NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+				BT_IO_OPT_DEST_BDADDR, &dev->dst,
+				BT_IO_OPT_INVALID);
+	if (!io) {
+		error("%s", err->message);
+		g_error_free(err);
+		return -EIO;
 	}
 
+	g_io_channel_unref(io);
+
 	headset_set_state(dev, HEADSET_STATE_PLAY_IN_PROGRESS);
 
 	pending_connect_init(hs, HEADSET_STATE_PLAYING);
@@ -656,7 +670,8 @@ int telephony_event_reporting_rsp(void *telephony_device, cme_error_t err)
 	if (hs->state != HEADSET_STATE_CONNECT_IN_PROGRESS)
 		return 0;
 
-	if (ag.features & AG_FEATURE_THREE_WAY_CALLING)
+	if (hs->hf_features & HF_FEATURE_CALL_WAITING_AND_3WAY &&
+			ag.features & AG_FEATURE_THREE_WAY_CALLING)
 		return 0;
 
 	hfp_slc_complete(device);
@@ -727,9 +742,15 @@ static int call_hold(struct audio_device *dev, const char *buf)
 	return 0;
 }
 
-static int button_press(struct audio_device *device, const char *buf)
+int telephony_key_press_rsp(void *telephony_device, cme_error_t err)
 {
-	struct headset *hs = device->headset;
+	return telephony_generic_rsp(telephony_device, err);
+}
+
+static int key_press(struct audio_device *device, const char *buf)
+{
+	if (strlen(buf) < 9)
+		return -EINVAL;
 
 	g_dbus_emit_signal(device->conn, device->path,
 			AUDIO_HEADSET_INTERFACE, "AnswerRequested",
@@ -740,7 +761,9 @@ static int button_press(struct audio_device *device, const char *buf)
 		ag.ring_timer = 0;
 	}
 
-	return headset_send(hs, "\r\nOK\r\n");
+	telephony_key_press_req(device, &buf[8]);
+
+	return 0;
 }
 
 int telephony_answer_call_rsp(void *telephony_device, cme_error_t err)
@@ -990,7 +1013,7 @@ static int call_waiting_notify(struct audio_device *device, const char *buf)
 
 	if (buf[8] == '1') {
 		hs->cwa_enabled = TRUE;
-		debug("Call waiting notifiaction enabled for headset %p", hs);
+		debug("Call waiting notification enabled for headset %p", hs);
 	} else {
 		hs->cwa_enabled = FALSE;
 		debug("Call waiting notification disabled for headset %p", hs);
@@ -1076,7 +1099,7 @@ static struct event event_callbacks[] = {
 	{ "AT+CMER", event_reporting },
 	{ "AT+CHLD", call_hold },
 	{ "AT+CHUP", terminate_call },
-	{ "AT+CKPD", button_press },
+	{ "AT+CKPD", key_press },
 	{ "AT+CLIP", cli_notification },
 	{ "AT+BTRH", response_and_hold },
 	{ "AT+BLDN", last_dialed_number },
@@ -1111,7 +1134,7 @@ static void close_sco(struct audio_device *device)
 	if (hs->sco) {
 		g_source_remove(hs->sco_id);
 		hs->sco_id = 0;
-		g_io_channel_close(hs->sco);
+		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 		hs->sco = NULL;
 	}
@@ -1122,11 +1145,8 @@ static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
 {
 	struct headset *hs;
 	unsigned char buf[BUF_SIZE];
-	char *cr;
 	gsize bytes_read = 0;
 	gsize free_space;
-	int err;
-	off_t cmd_len;
 
 	if (cond & G_IO_NVAL)
 		return FALSE;
@@ -1156,27 +1176,39 @@ static gboolean rfcomm_io_cb(GIOChannel *chan, GIOCondition cond,
 	 * functions */
 	hs->buf[hs->data_start + hs->data_length] = '\0';
 
-	cr = strchr(&hs->buf[hs->data_start], '\r');
-	if (!cr)
-		return TRUE;
+	while (hs->data_length > 0) {
+		char *cr;
+		int err;
+		off_t cmd_len;
 
-	cmd_len = 1 + (off_t) cr - (off_t) &hs->buf[hs->data_start];
-	*cr = '\0';
+		cr = strchr(&hs->buf[hs->data_start], '\r');
+		if (!cr)
+			break;
 
-	err = handle_event(device, &hs->buf[hs->data_start]);
-	if (err == -EINVAL) {
-		error("Badly formated or unrecognized command: %s",
-				&hs->buf[hs->data_start]);
-		err = headset_send(hs, "\r\nERROR\r\n");
-	} else if (err < 0)
-		error("Error handling command %s: %s (%d)",
-			&hs->buf[hs->data_start], strerror(-err), -err);
+		cmd_len = 1 + (off_t) cr - (off_t) &hs->buf[hs->data_start];
+		*cr = '\0';
 
-	hs->data_start += cmd_len;
-	hs->data_length -= cmd_len;
+		if (cmd_len > 1)
+			err = handle_event(device, &hs->buf[hs->data_start]);
+		else
+			/* Silently skip empty commands */
+			err = 0;
 
-	if (!hs->data_length)
-		hs->data_start = 0;
+		if (err == -EINVAL) {
+			error("Badly formated or unrecognized command: %s",
+					&hs->buf[hs->data_start]);
+			err = headset_send(hs, "\r\nERROR\r\n");
+		} else if (err < 0)
+			error("Error handling command %s: %s (%d)",
+						&hs->buf[hs->data_start],
+						strerror(-err), -err);
+
+		hs->data_start += cmd_len;
+		hs->data_length -= cmd_len;
+
+		if (!hs->data_length)
+			hs->data_start = 0;
+	}
 
 	return TRUE;
 
@@ -1203,22 +1235,37 @@ static gboolean sco_cb(GIOChannel *chan, GIOCondition cond,
 	return FALSE;
 }
 
-static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-			const bdaddr_t *dst, gpointer user_data)
+void headset_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 {
 	struct audio_device *dev = user_data;
 	struct headset *hs = dev->headset;
 	struct pending_connect *p = hs->pending;
 	char hs_address[18];
 
-	if (err < 0) {
-		error("connect(): %s (%d)", strerror(-err), -err);
+	if (err) {
+		error("%s", err->message);
 		goto failed;
 	}
 
+	/* For HFP telephony isn't ready just disconnect */
+	if (hs->hfp_active && !ag.telephony_ready) {
+		error("Unable to accept HFP connection since the telephony "
+				"subsystem isn't initialized");
+		goto failed;
+	}
+
+	if (hs->tmp_rfcomm) {
+		hs->rfcomm = hs->tmp_rfcomm;
+		hs->tmp_rfcomm = NULL;
+	} else
+		hs->rfcomm = g_io_channel_ref(chan);
+
 	ba2str(&dev->dst, hs_address);
-	hs->rfcomm = chan;
-	p->io = NULL;
+
+	if (p)
+		p->io = NULL;
+	else
+		hs->auto_dc = FALSE;
 
 	if (server_is_enabled(&dev->src, HANDSFREE_SVCLASS_ID) &&
 			hs->hfp_handle != 0)
@@ -1237,14 +1284,14 @@ static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 
 	headset_set_state(dev, HEADSET_STATE_CONNECTED);
 
-	if (p->target_state == HEADSET_STATE_PLAYING) {
+	if (p && p->target_state == HEADSET_STATE_PLAYING) {
 		p->err = sco_connect(dev, NULL, NULL, NULL);
 		if (p->err < 0)
 			goto failed;
 		return;
 	}
 
-	if (p->msg) {
+	if (p && p->msg) {
 		DBusMessage *reply = dbus_message_new_method_return(p->msg);
 		g_dbus_send_message(dev->conn, reply);
 	}
@@ -1254,7 +1301,7 @@ static void rfcomm_connect_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 	return;
 
 failed:
-	if (p->msg)
+	if (p && p->msg)
 		error_connection_attempt_failed(dev->conn, p->msg, p->err);
 	pending_connect_finalize(dev);
 	if (hs->rfcomm)
@@ -1379,7 +1426,8 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 {
 	struct headset *hs = dev->headset;
 	char address[18];
-	int err;
+	GError *err = NULL;
+	GIOChannel *io;
 
 	if (hs->rfcomm_ch < 0)
 		return get_records(dev, cb, user_data, cb_id);
@@ -1389,13 +1437,20 @@ static int rfcomm_connect(struct audio_device *dev, headset_stream_cb_t cb,
 	debug("%s: Connecting to %s channel %d", dev->path, address,
 		hs->rfcomm_ch);
 
-	err = bt_rfcomm_connect(&dev->src, &dev->dst, hs->rfcomm_ch,
-			rfcomm_connect_cb, dev);
-	if (err < 0) {
-		error("connect() failed: %s (%d)", strerror(-err), -err);
-		return err;
+	io = bt_io_connect(BT_IO_RFCOMM, headset_connect_cb, dev,
+				NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &dev->src,
+				BT_IO_OPT_DEST_BDADDR, &dev->dst,
+				BT_IO_OPT_CHANNEL, hs->rfcomm_ch,
+				BT_IO_OPT_INVALID);
+	if (!io) {
+		error("%s", err->message);
+		g_error_free(err);
+		return -EIO;
 	}
 
+	g_io_channel_unref(io);
+
 	headset_set_state(dev, HEADSET_STATE_CONNECT_IN_PROGRESS);
 
 	pending_connect_init(hs, HEADSET_STATE_CONNECTED);
@@ -1973,12 +2028,12 @@ static void headset_free(struct audio_device *dev)
 	}
 
 	if (hs->sco) {
-		g_io_channel_close(hs->sco);
+		g_io_channel_shutdown(hs->sco, TRUE, NULL);
 		g_io_channel_unref(hs->sco);
 	}
 
 	if (hs->rfcomm) {
-		g_io_channel_close(hs->rfcomm);
+		g_io_channel_shutdown(hs->rfcomm, TRUE, NULL);
 		g_io_channel_unref(hs->rfcomm);
 	}
 
@@ -2227,10 +2282,11 @@ unsigned int headset_suspend_stream(struct audio_device *dev,
 		hs->dc_timer = 0;
 	}
 
-	close_sco(dev);
+	headset_set_state(dev, HEADSET_STATE_CONNECTED);
 
 	id = connect_cb_new(hs, HEADSET_STATE_CONNECTED, cb, user_data);
 	g_idle_add((GSourceFunc) dummy_connect_complete, dev);
+
 	return id;
 }
 
@@ -2248,13 +2304,23 @@ void set_hfp_active(struct audio_device *dev, gboolean active)
 	hs->hfp_active = active;
 }
 
+GIOChannel *headset_get_rfcomm(struct audio_device *dev)
+{
+	struct headset *hs = dev->headset;
+
+	return hs->tmp_rfcomm;
+}
+
 int headset_connect_rfcomm(struct audio_device *dev, GIOChannel *io)
 {
 	struct headset *hs = dev->headset;
 
-	hs->tmp_rfcomm = io;
+	if (hs->tmp_rfcomm)
+		return -EALREADY;
 
-	return hs->tmp_rfcomm ? 0 : -EINVAL;
+	hs->tmp_rfcomm = g_io_channel_ref(io);
+
+	return 0;
 }
 
 int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
@@ -2264,7 +2330,7 @@ int headset_connect_sco(struct audio_device *dev, GIOChannel *io)
 	if (hs->sco)
 		return -EISCONN;
 
-	hs->sco = io;
+	hs->sco = g_io_channel_ref(io);
 
 	if (hs->pending_ring) {
 		ring_timer_cb(NULL);
@@ -2283,7 +2349,7 @@ static int headset_close_rfcomm(struct audio_device *dev)
 	GIOChannel *rfcomm = hs->tmp_rfcomm ? hs->tmp_rfcomm : hs->rfcomm;
 
 	if (rfcomm) {
-		g_io_channel_close(rfcomm);
+		g_io_channel_shutdown(rfcomm, TRUE, NULL);
 		g_io_channel_unref(rfcomm);
 		hs->tmp_rfcomm = NULL;
 		hs->rfcomm = NULL;
@@ -2297,32 +2363,6 @@ static int headset_close_rfcomm(struct audio_device *dev)
 	return 0;
 }
 
-void headset_set_authorized(struct audio_device *dev)
-{
-	struct headset *hs = dev->headset;
-
-	/* For HFP telephony isn't ready just disconnect */
-	if (hs->hfp_active && !ag.telephony_ready) {
-		error("Unable to accept HFP connection since the telephony "
-				"subsystem isn't initialized");
-		headset_set_state(dev, HEADSET_STATE_DISCONNECTED);
-		return;
-	}
-
-	hs->rfcomm = hs->tmp_rfcomm;
-	hs->tmp_rfcomm = NULL;
-
-	g_io_add_watch(hs->rfcomm,
-			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-			(GIOFunc) rfcomm_io_cb, dev);
-
-	hs->auto_dc = FALSE;
-
-	/* For HSP (no special SLC setup) move to CONNECTED state */
-	if (!hs->hfp_active)
-		headset_set_state(dev, HEADSET_STATE_CONNECTED);
-}
-
 void headset_set_state(struct audio_device *dev, headset_state_t state)
 {
 	struct headset *hs = dev->headset;
@@ -2491,6 +2531,11 @@ gboolean headset_get_nrec(struct audio_device *dev)
 	return hs->nrec;
 }
 
+gboolean headset_get_sco_hci(struct audio_device *dev)
+{
+	return sco_hci;
+}
+
 int telephony_event_ind(int index)
 {
 	if (!active_devices)
