@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -120,6 +120,39 @@ int read_discoverable_timeout(const char *src, int *timeout)
 	return 0;
 }
 
+int write_pairable_timeout(bdaddr_t *bdaddr, int timeout)
+{
+	char filename[PATH_MAX + 1], str[32];
+
+	snprintf(str, sizeof(str), "%d", timeout);
+
+	create_filename(filename, PATH_MAX, bdaddr, "config");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	return textfile_put(filename, "pairto", str);
+}
+
+int read_pairable_timeout(const char *src, int *timeout)
+{
+	char filename[PATH_MAX + 1], *str;
+
+	create_name(filename, PATH_MAX, STORAGEDIR, src, "config");
+
+	str = textfile_get(filename, "pairto");
+	if (!str)
+		return -ENOENT;
+
+	if (sscanf(str, "%d", timeout) != 1) {
+		free(str);
+		return -ENOENT;
+	}
+
+	free(str);
+
+	return 0;
+}
+
 int write_device_mode(bdaddr_t *bdaddr, const char *mode)
 {
 	char filename[PATH_MAX + 1];
@@ -339,6 +372,37 @@ int write_remote_eir(bdaddr_t *local, bdaddr_t *peer, uint8_t *data)
 	return textfile_put(filename, addr, str);
 }
 
+int read_remote_eir(bdaddr_t *local, bdaddr_t *peer, uint8_t *data)
+{
+	char filename[PATH_MAX + 1], addr[18], *str;
+	int i;
+
+	create_filename(filename, PATH_MAX, local, "eir");
+
+	ba2str(peer, addr);
+
+	str = textfile_get(filename, addr);
+	if (!str)
+		return -ENOENT;
+
+	if (!data) {
+		free(str);
+		return 0;
+	}
+
+	if (strlen(str) < 480) {
+		free(str);
+		return -EIO;
+	}
+
+	for (i = 0; i < 240; i++)
+		sscanf(str + (i * 2), "%02hhX", &data[i]);
+
+	free(str);
+
+	return 0;
+}
+
 int write_l2cap_info(bdaddr_t *local, bdaddr_t *peer,
 			uint16_t mtu_result, uint16_t mtu,
 			uint16_t mask_result, uint32_t mask)
@@ -403,13 +467,13 @@ int read_l2cap_info(bdaddr_t *local, bdaddr_t *peer,
 	return 0;
 }
 
-int write_version_info(bdaddr_t *local, bdaddr_t *peer, uint16_t err,
+int write_version_info(bdaddr_t *local, bdaddr_t *peer, uint16_t manufacturer,
 					uint8_t lmp_ver, uint16_t lmp_subver)
 {
 	char filename[PATH_MAX + 1], addr[18], str[16];
 
 	memset(str, 0, sizeof(str));
-	sprintf(str, "%d %d %d", err, lmp_ver, lmp_subver);
+	sprintf(str, "%d %d %d", manufacturer, lmp_ver, lmp_subver);
 
 	create_filename(filename, PATH_MAX, local, "manufacturers");
 
@@ -783,8 +847,8 @@ sdp_record_t *record_from_string(const gchar *str)
 
 	tmp[2] = 0;
 	for (i = 0; i < size; i++) {
-		 memcpy(tmp, str + (i * 2), 2);
-		 pdata[i] = (uint8_t) strtol(tmp, NULL, 16);
+		memcpy(tmp, str + (i * 2), 2);
+		pdata[i] = (uint8_t) strtol(tmp, NULL, 16);
 	}
 
 	rec = sdp_extract_pdu(pdata, size, &len);
@@ -1004,8 +1068,8 @@ int read_device_id(const gchar *srcaddr, const gchar *dstaddr,
 	if (err) {
 		/* FIXME: We should try EIR data if we have it, too */
 
-		/* If we don't have the data, we don't want to go
-		   through the above search every time. */
+		/* If we don't have the data, we don't want to go through the
+		 * above search every time. */
 		lsource = 0xffff;
 		lvendor = 0x0000;
 		lproduct = 0x0000;
@@ -1028,3 +1092,33 @@ int read_device_id(const gchar *srcaddr, const gchar *dstaddr,
 
 	return 0;
 }
+
+int write_device_pairable(bdaddr_t *bdaddr, gboolean mode)
+{
+	char filename[PATH_MAX + 1];
+
+	create_filename(filename, PATH_MAX, bdaddr, "config");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	return textfile_put(filename, "pairable", mode ? "yes" : "no");
+}
+
+int read_device_pairable(bdaddr_t *bdaddr, gboolean *mode)
+{
+	char filename[PATH_MAX + 1], *str;
+
+	create_filename(filename, PATH_MAX, bdaddr, "config");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	str = textfile_get(filename, "pairable");
+	if (!str)
+		return -ENOENT;
+
+	*mode = strcmp(str, "yes") == 0 ? TRUE : FALSE;
+
+	free(str);
+
+	return 0;
+}
