@@ -3,7 +3,7 @@
  *  BlueZ - Bluetooth protocol stack for Linux
  *
  *  Copyright (C) 2006-2007  Nokia Corporation
- *  Copyright (C) 2004-2008  Marcel Holtmann <marcel@holtmann.org>
+ *  Copyright (C) 2004-2009  Marcel Holtmann <marcel@holtmann.org>
  *
  *
  *  This program is free software; you can redistribute it and/or modify
@@ -500,8 +500,8 @@ static int avdtp_get_caps(int sk, int seid)
 	if (ret < 0)
 		return ret;
 
-	if (ret < (sizeof(struct getcap_resp) + 4 +
-			sizeof(struct avdtp_media_codec_capability))) {
+	if (ret < ((int) sizeof(struct getcap_resp) + 4 +
+			(int) sizeof(struct avdtp_media_codec_capability))) {
 		printf("Invalid capabilities\n");
 		return -1;
 	}
